/* GRJ, Apr/May 2020
 */


package pomocov;

import beast.core.Distribution;
import beast.core.Input;
import beast.core.State;

import java.util.List;


import java.util.Random;

import static pomocov.utils.TransmissionDistributions.*;

/*
calculates logPrior for transmission tree
 */

public class TreePoMoCoVPrior  extends Distribution {

    final public Input<TreePoMo> treeInput = new Input<>("tree",
            "phylogenetic transmission PoMo tree with sequence data in the leafs and internal nodes",
            Input.Validate.REQUIRED);

final public Input<Double> unobservedCountPriorMeanInput = new Input<>("unobservedCountPriorMean",
        "The prior mean of the number of unobserved hosts between an observed host and its observed source host (default 1.0)",
        1.0);



    @Override
    public List<String> getArguments() {
        return null;
    }

    @Override
    public List<String> getConditions() {
        return null;
    }

    @Override
    public void sample(State state, Random random) {

    }

    @Override
    public void initAndValidate() {
        double ucMean = unobservedCountPriorMeanInput.get();
        if (ucMean < 0.0) {
            System.err.println("unobservedCountPriorMean must be non-negative");
            System.exit(1);
        }
    }

    @Override
    public double calculateLogP() {
        logP = 0.0;
        if (treeInput.get().getUnobservedHosts()) {
            double ucMean = unobservedCountPriorMeanInput.get();
            logP = calculateTreeLogPrior(treeInput.get(), ucMean);
        }
        return logP;
    }



    private double calculateTreeLogPrior(TreePoMo tree, double ucMean) {
        double logPrior = 0.0;
        NodePoMo[] nodes = tree.getNodesAsArray();
        for (NodePoMo node : nodes) {
            double h = node.getHeight();
            if (node.getParent() != null) {
                double g = node.getOriginHeight();
                int u = node.getUnobservedCount();
                logPrior += betweenHostsLogPrior(u, g, h);
                logPrior += unobservedCountLogPrior(u, ucMean);
            }
            for (NodePoMo ch : node.getChildren()) {
                logPrior += withinHostLogPrior(h - ch.getHeight());
            }
        }
        return logPrior;
    }






}
