
/* GRJ, Apr/May 2020
 */


package pomocov.utils;

import java.util.Formatter;
import java.util.Locale;

public class DebugTune {

    public static String padStringTo(String s, int n) {
        String ss = new String(s);
        int add = n - ss.length();
        for (int i = 0; i < add; i++) {
            ss += " ";
        }
        return ss;
    }


    public static String nonnegIn8Chars(double x) {
        StringBuilder s = new StringBuilder();
        Formatter formatter = new Formatter(s, Locale.US);
        if (x < 0) {
            formatter.format(" neg    ");
        } else if (x <= 1e-10) {
            formatter.format(" 0      ");
        } else if (x < 1e-3) {
            formatter.format("%8.2e", x);
        } else if (x < 9.999) {
            formatter.format("%8.5f", x);
        } else if (x < 99.99) {
            formatter.format("%8.4f", x);
        } else if (x < 999.9) {
            formatter.format("%8.3f", x);
        } else if (x < 9999) {
            formatter.format("%8.2f", x);
        } else {
            formatter.format("%8.0f", x);
        }
        return s.toString();
    }


    public static String floatIn9Chars(double x) {
        StringBuilder s = new StringBuilder();
        Formatter formatter = new Formatter(s, Locale.US);
        String sign = " ";
        if (x < 0) {
            sign = "-";
            x = -x;
        }
        if (x <= 1e-10) {
            formatter.format("%2s0      ", sign);
        } else if (x < 1e-3) {
            formatter.format("%2s%7.2e", sign, x);
        } else if (x < 9.999) {
            formatter.format("%2s%7.5f", sign, x);
        } else if (x < 99.99) {
            formatter.format("%2s%7.4f", sign, x);
        } else if (x < 999.9) {
            formatter.format("%2s%7.3f", sign, x);
        } else if (x < 9999) {
            formatter.format("%2s%7.2f", sign, x);
        } else {
            formatter.format("%2s%7.0f", sign, x);
        }
        return s.toString();
    }

    public static String floatIn15Chars(double x) {
        StringBuilder s = new StringBuilder();
        Formatter formatter = new Formatter(s, Locale.US);
        formatter.format("%15.8f", x);
        return s.toString();
    }


    public static void errOutBriefNonNeg1DMatrix(String name, double [] M) {
        int n = (int) Math.floor(0.5 + Math.sqrt(M.length));
        assert n*n == M.length;
        System.err.println(name);
        for (int row = 0; row < n; row++) {
            for (int col = 0; col < n; col++) {
                System.err.print(nonnegIn8Chars(M[n * row + col]) + " ");
            }
            System.err.println("");
        }
        System.err.println("");
    }

    public static void errOutNonNeg1DMatrix(String name, double [] M) {
        int n = (int) Math.floor(0.5 + Math.sqrt(M.length));
        assert n*n == M.length;
        System.err.println(name);
        for (int row = 0; row < n; row++) {
            for (int col = 0; col < n; col++) {
                System.err.print(floatIn15Chars(M[n * row + col]) + " ");
            }
            System.err.println("");
        }
        System.err.println("");
    }


    public static void errOutSigned2DMatrix(String name, double [][] M) {
        int n = M.length;
        assert n > 0;
        assert n == M[0].length;
        System.err.println(name);
        for (int row = 0; row < n; row++) {
            for (int col = 0; col < n; col++) {
                System.err.print(floatIn9Chars(M[row][col]));
            }
            System.err.println("");
        }
        System.err.println("");
    }


}
